"""
Interface to git's word-level diff (git diff --color-word).
Requires a system install of git.
"""
import tempfile
from subprocess import run


def wdiff(st1: str, st2: str, *, return_exit_code=False, line_offset=0):
    """
    Interface to git's word-level diff.
    """
    with tempfile.NamedTemporaryFile(mode="w") as sf1, tempfile.NamedTemporaryFile(
        mode="w"
    ) as sf2:
        sf1.write("\n" * line_offset)
        sf1.write(st1)
        sf2.write("\n" * line_offset)
        sf2.write(st2)
        sf1.flush()
        sf2.flush()
        result = run(
            [
                "git",
                "--no-pager",
                "diff",
                "--ignore-space-at-eol",
                "--no-index",
                "--exit-code",
                "--color-words",
                sf1.name,
                sf2.name,
            ],
            capture_output=True,
            check=False,
        )
        if return_exit_code:
            return result.stdout, result.returncode
        else:
            return result.stdout



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from subprocess import PIPE, Popen, run


project = "catleg"
copyright = "2023, the catala contributors"
author = "the catala contributors"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# version
pipe = Popen("hatch version", stdout=PIPE, shell=True)
version = pipe.stdout.read().decode("utf8")

# generate CLI reference
# (this command implies that `make html` is executed from the `docs`
# directory -- this could be made more robust?)
run(
    [
        "typer",
        "../src/catleg/catleg.py",
        "utils",
        "docs",
        "--name",
        "catleg",
        "--output",
        "cli_reference.md",
    ],
    check=True,
)



import sys
import warnings
from pathlib import Path
from typing import TextIO

from catleg.git_diff import wdiff

from catleg.parse_catala_markdown import parse_catala_file
from catleg.query import get_backend


async def find_changes(f: TextIO, *, file_path: Path | None = None):
    # parse articles from file
    articles = parse_catala_file(f, file_path=file_path)

    # fetch articles' reference text
    # compute diff
    # display diff
    back = get_backend("legifrance")
    ref_articles = await back.articles([article.id.upper() for article in articles])

    diffcnt = 0
    for article, ref_article in zip(articles, ref_articles):
        if ref_article is None:
            warnings.warn(f"Could not retrieve article '{article.id}'")
            continue

        diff, retcode = wdiff(
            _reformat(article.text),
            _reformat(_escape_ref_text(ref_article.text_and_nota())),
            return_exit_code=True,
            line_offset=article.start_line,
        )
        if retcode != 0:
            print(article.id)
            print(f"{article.file_path or 'UNKNOWN_FILE'}:{article.start_line}")
            sys.stdout.buffer.write(diff)
            diffcnt += 1
    if diffcnt > 0:
        sys.stdout.flush()
        print(
            f"Found {diffcnt} articles with diffs (out of {len(articles)} articles)",
            file=sys.stderr,
        )
    # (ci mode : error code != 0 if any diff?)


def _reformat(paragraph: str):
    """
    Catala has a 80-char line limit, so law texts will often be manually reformatted.
    We attempt to remove extra line breaks before comparison.
    """
    paragraph = paragraph.replace("\n", " ")
    while "  " in paragraph:
        paragraph = paragraph.replace("  ", " ")
    return paragraph.strip()


def _escape_ref_text(paragraph: str):
    """
    Escape markdown special chars from reference text.
    Since we use the "raw text" version of the Legifrance article,
    we escape brackets and asteriks unconditionnally.
    """
    return paragraph.replace("[", r"\[").replace("]", r"\]").replace("*", r"\*")



from catleg.config import settings


def set_basic_loglevel():
    """
    Utility for setting the log level as per config -- meant to be used
    within CLI tools only.

    To use, set CATLEG_LOG_LEVEL=INFO in the environment
    or log_level in the .catleg.toml configuration file
    """
    log_level = settings.get("log_level")
    if log_level is not None:
        import logging

        logging.basicConfig(level=log_level.upper())



from catleg.query import _article_from_legifrance_reply
from catleg.skeleton import _formatted_atricle

from .test_legifrance_queries import _json_from_test_file


# Regression test for https://github.com/CatalaLang/catleg/issues/71
def test_no_article_renumbering():
    article_json = _json_from_test_file("LEGIARTI000044983201.json")
    article = _article_from_legifrance_reply(article_json)
    formatted_article_md = _formatted_atricle(article)
    assert "1. Le bénéfice ou revenu imposable est constitué" in formatted_article_md
    assert "2. Le revenu global net annuel" in formatted_article_md
    assert (
        "3. Le bénéfice ou revenu net de chacune des catégories" in formatted_article_md
    )



import asyncio
import os
from json import load

import pytest
from catleg.query import (
    _article_from_legifrance_reply,
    _get_legifrance_credentials,
    get_backend,
)


def _json_from_test_file(fname):
    path_to_current_file = os.path.realpath(__file__)
    current_directory = os.path.dirname(path_to_current_file)
    file_path = os.path.join(current_directory, fname)
    with open(file_path, encoding="utf-8") as f:
        json_contents = load(f)
    return json_contents


@pytest.mark.skipif(
    _get_legifrance_credentials(raise_if_missing=False)[0] is None,
    reason="this test requires legifrance credentials",
)
def test_query_article():
    back = get_backend("legifrance")
    article = asyncio.run(back.article("LEGIARTI000038814944"))
    assert "logement" in article.text


@pytest.mark.skipif(
    _get_legifrance_credentials(raise_if_missing=False)[0] is None,
    reason="this test requires legifrance credentials",
)
def test_query_article_2():
    back = get_backend("legifrance")
    article = asyncio.run(back.article("CETATEXT000035260342"))
    assert article is not None


@pytest.mark.skipif(
    _get_legifrance_credentials(raise_if_missing=False)[0] is None,
    reason="this test requires legifrance credentials",
)
def test_query_articles():
    back = get_backend("legifrance")
    art1, art2 = asyncio.run(
        back.articles(["LEGIARTI000038814944", "LEGIARTI000038814948"])
    )
    assert "logement" in art1.text
    assert "pacte" in art2.text


@pytest.mark.skipif(
    _get_legifrance_credentials(raise_if_missing=False)[0] is None,
    reason="this test requires legifrance credentials",
)
@pytest.mark.parametrize("page_size", [20, 100])
def test_query_codes(page_size):
    back = get_backend("legifrance")
    codes, nb_results = asyncio.run(back._list_codes(page_size))
    assert nb_results == len(codes)


def test_no_extraneous_nota():
    article = _json_from_test_file("LEGIARTI000038814944.json")
    res = _article_from_legifrance_reply(article)
    assert "NOTA" not in res.to_markdown()
    assert "NOTA" not in res.text_and_nota()


def test_nota_format():
    article = _json_from_test_file("LEGIARTI000046790860.json")
    res = _article_from_legifrance_reply(article)
    assert "\n\nNOTA :\n\nConformément à l'article 89" in res.to_markdown()
    assert "\n\nNOTA :\n\nConformément à l'article 89" in res.text_and_nota()


def test_keep_newlines():
    article = _json_from_test_file("LEGIARTI000046790860.json")
    res = _article_from_legifrance_reply(article)
    assert "\n" in res.to_markdown()


def test_null_notas():
    article = _json_from_test_file("JORFARTI000046186676.json")
    res = _article_from_legifrance_reply(article)
    assert res.nota == ""
    assert res.nota_html == ""


def test_strip_links_from_markdown():
    article = _json_from_test_file("LEGIARTI000006302217.json")
    res = _article_from_legifrance_reply(article)
    assert "affichCodeArticle.do" not in res.to_markdown()


def test_expiry():
    article = _json_from_test_file("LEGIARTI000046790860.json")
    res = _article_from_legifrance_reply(article)
    assert res.is_open_ended

    # Conseil d'État texts do not expire
    ceta_json_article = _json_from_test_file("CETATEXT000035260342.json")
    ceta_article = _article_from_legifrance_reply(ceta_json_article)
    assert ceta_article.is_open_ended



from dynaconf import Dynaconf  # type:ignore

settings = Dynaconf(
    envvar_prefix="CATLEG",
    settings_files=[".catleg.toml", ".catleg_secrets.toml"],
)

# `envvar_prefix` = export envvars with `export DYNACONF_FOO=bar`.
# `settings_files` = Load these files in the order.



import asyncio
import json
from pathlib import Path

import typer

from catleg.check_expiry import check_expiry as expiry
from catleg.cli_util import set_basic_loglevel
from catleg.find_changes import find_changes
from catleg.query import get_backend
from catleg.skeleton import markdown_skeleton

app = typer.Typer(add_completion=False)
# legifrance-specific commands (query legifrance API and return
# raw JSON)
lf = typer.Typer()
app.add_typer(lf, name="lf", help="Commands for querying the raw Legifrance API")


@app.command()
def diff(file: Path):
    """
    Show differences between each article in a catala file and
    a reference version.
    """
    with open(file) as f:
        asyncio.run(find_changes(f, file_path=file))


@app.command()
def check_expiry(file: Path):
    """
    Check articles in a catala file for expiry.
    """
    with open(file) as f:
        retcode = asyncio.run(expiry(f, file_path=file))
        raise typer.Exit(retcode)


@app.command()
def skeleton(textid: str, sectionid: str):
    """
    Output a given section of a law text.
    """
    skel = asyncio.run(markdown_skeleton(textid, sectionid))
    print(skel)


@lf.command()
def article(article_id: str):
    """Retrieve an article from Legifrance"""
    back = get_backend("legifrance")
    print(
        json.dumps(
            asyncio.run(back.query_article_legi(article_id)),
            indent=2,
            ensure_ascii=False,
        )
    )


@lf.command()
def codes():
    """
    Retrieve a list of available codes.
    """
    back = get_backend("legifrance")
    print(json.dumps(asyncio.run(back.list_codes()), indent=2, ensure_ascii=False))


@lf.command()
def toc(code: str):
    """
    Retrieve the table of contents for a given code.
    """
    back = get_backend("legifrance")
    print(json.dumps(asyncio.run(back.code_toc(code)), indent=2, ensure_ascii=False))


def main():
    set_basic_loglevel()

    app()



from pathlib import Path
from typing import cast, TextIO

from markdown_it import MarkdownIt
from markdown_it.tree import SyntaxTreeNode
from mdformat.renderer import MDRenderer
from more_itertools import sliding_window

from catleg.law_text_fr import CatalaFileArticle, find_id_in_string
from catleg.markdown_it.heading_extension import replace_heading_rule


def parse_catala_file(
    f: TextIO, file_path: Path | None = None
) -> list[CatalaFileArticle]:
    """
    Given a catala file, return a list of articles
    """
    md = _make_markdown_parser()
    tokens = md.parse(f.read())
    tree = SyntaxTreeNode(tokens)
    articles = _parse_catala_doc(tree, file_path=file_path)
    return articles


def _make_markdown_parser() -> MarkdownIt:
    """
    Return a markdown parser suitable for Catala files
    (this currently is a CommonMark parser with an extension that
    increases the number of possible heading levels)
    """
    return replace_heading_rule(MarkdownIt("commonmark"))


def _parse_catala_doc(
    tree: SyntaxTreeNode, file_path: Path | None = None
) -> list[CatalaFileArticle]:
    articles: list[CatalaFileArticle] = []
    windowed_tree = sliding_window(tree.walk(), 3)
    renderer = MDRenderer()
    for window in windowed_tree:
        match [elem.type for elem in window]:
            case ["heading", "inline", "text"]:
                if type_id := find_id_in_string(window[-1].content):
                    typ, id = type_id
                    text = []
                    curr_elem = window[0]
                    is_archive = "[archive]" in window[-1].content

                    # start line of the first text block
                    inline_map = window[1].map
                    assert inline_map is not None
                    start_line: int = cast(tuple[int, int], inline_map)[0]

                    while curr_elem.next_sibling is not None:
                        if curr_elem.next_sibling.type == "heading":
                            break
                        # skip code blocks, retain all other elements
                        if curr_elem.next_sibling.type != "fence":
                            text.append(
                                renderer.render(
                                    curr_elem.next_sibling.to_tokens(),
                                    options={},
                                    env={},
                                )
                            )
                        curr_elem = curr_elem.next_sibling

                    articles.append(
                        CatalaFileArticle(
                            type=typ,
                            id=id,
                            text=" ".join(text),
                            start_line=start_line,
                            file_path=file_path,
                            is_archive=is_archive,
                        )
                    )

    return articles


if __name__ == "__main__":
    import sys

    with open(sys.argv[1]) as f:
        doc = parse_catala_file(f)
        print(doc)



from itertools import dropwhile

import mdformat

from catleg.query import get_backend


async def markdown_skeleton(textid: str, sectionid: str) -> str:
    """
    Return a skeleton (formatted law text section)
    """
    if sectionid[:8].upper() != "LEGISCTA":
        raise ValueError("Expected section identifier (should start with 'SCTA')")

    back = get_backend("legifrance")
    toc = await back.code_toc(textid)

    nodes = dropwhile(
        lambda node_level: node_level[0]["cid"] != sectionid, _preorder(toc)
    )

    # some existence checking is needed here
    root, root_level = next(nodes)
    parts = []

    for node, level in _preorder(root, root_level):
        if node["id"][:8] != "LEGISCTA":
            # If it is not a section, then it is an article
            parts.append(f"{'#' * (level + 1)} Article {node['num']} | {node['id']}")
            article = await back.article(node["id"])
            parts.append(_formatted_atricle(article))
        else:
            parts.append(f"{'#' * level} {node['title']}")

    return "\n\n".join(parts)


def _preorder(node, level=1):
    """Preorder traversal of articles and sections"""
    yield node, level
    for article in node["articles"]:
        yield article, level
    for section in node["sections"]:
        yield from _preorder(section, level + 1)


def _formatted_atricle(article):
    return mdformat.text(article.to_markdown(), options={"wrap": 80, "number": True})



from io import StringIO

from catleg.law_text_fr import ArticleType, find_id_in_string, parse_article_id
from catleg.parse_catala_markdown import _make_markdown_parser, parse_catala_file
from mdformat.renderer import MDRenderer


def test_parse_article_id():
    typ, id = parse_article_id("legiarti000038814944")
    assert typ == ArticleType.LEGIARTI
    assert id == "legiarti000038814944"


def test_find_id_in_string():
    assert find_id_in_string("####### Article R821-2 | LEGIARTI000038879021") == (
        ArticleType.LEGIARTI,
        "LEGIARTI000038879021",
    )
    assert find_id_in_string("###### Article L841-1 | LEGIARTI000038814864") == (
        ArticleType.LEGIARTI,
        "LEGIARTI000038814864",
    )
    assert find_id_in_string("This definitely does not match") is None


catala_text = """
###### Article L822-2 | LEGIARTI000038814944

I.-Peuvent bénéficier d'une aide personnelle au logement :

1° Les personnes de nationalité française ;

2° Les personnes de nationalité étrangère remplissant les conditions
prévues par les deux premiers alinéas de l' article M. 512-2 du
code de la sécurité sociale .

```catala
champ d'application ÉligibilitéAidesPersonnelleLogement:
  définition condition_nationalité égal à
    selon demandeur.nationalité sous forme
    -- Française: vrai
    -- Étrangère de conditions:
      conditions.satisfait_conditions_l512_2_code_sécurité_sociale
```

II.-Parmi les personnes mentionnées au I, peuvent bénéficier d'une aide
personnelle au logement celles remplissant les conditions prévues par le
présent livre pour son attribution qui sont locataires, résidents en
logement-foyer ou qui accèdent à la propriété d'un local à usage exclusif
d'habitation et constituant leur résidence principale.

Les sous-locataires, sous les mêmes conditions, peuvent également en bénéficier.

```catala
champ d'application ÉligibilitéAidesPersonnelleLogement:
  étiquette l822_2 règle condition_logement_mode_occupation sous condition
    selon ménage.logement.mode_occupation sous forme
      -- Locataire: vrai
      -- RésidentLogementFoyer: vrai
      -- AccessionPropriétéLocalUsageExclusifHabitation:
        ménage.logement.résidence_principale
      -- SousLocataire: vrai
      -- LocationAccession: vrai # Justifié par L831-2, avec une
                                 # exception à venir
  conséquence rempli
```

######## Article L821-2 | LEGIARTI000038814974

Les aides personnelles au logement sont accordées au titre de
la résidence principale.

```catala
champ d'application ÉligibilitéAidesPersonnelleLogement:
  règle condition_logement_résidence_principale sous condition
    ménage.logement.résidence_principale
  conséquence rempli
```
"""


def test_parse_markdown():
    # Note that we deviate from the markdown spec as
    # 7 on more '#' signs are parsed as a markdown atx heading
    # (vanilla markdown has only up to 6 heading levels, mirroring
    # html)
    f = StringIO(catala_text)
    articles = parse_catala_file(f)
    assert len(articles) == 2
    ids = [article.id for article in articles]
    assert "LEGIARTI000038814944" in ids
    assert "LEGIARTI000038814974" in ids


def test_emit_markdown():
    md_parser = _make_markdown_parser()
    tok_list = md_parser.parse(catala_text)
    renderer = MDRenderer()
    md_result = renderer.render(tok_list, {"wrap": 80}, {})
    # Note that we deviate from the markdown spec as we may
    # emit more than 6 '#' characters for atx headings.
    assert "######## Article L821-2 | LEGIARTI000038814974" in md_result



from __future__ import annotations

import logging

from itertools import islice, repeat

from markdown_it import MarkdownIt
from markdown_it.common.utils import isSpace
from markdown_it.rules_block.state_block import StateBlock

"""
Arbitrary-level (but not configurable) heading extension for markdown-it-py.

This is needed because law texts go beyond 6 levels of headers, and we do want to
stay true to them without adding new syntax.

This file embeds code from the markdown-it-py project, licensed under the MIT license.

MIT License

Copyright (c) 2020 ExecutableBookProject

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

""" Atex heading (#, ##, ...) """

LOGGER = logging.getLogger(__name__)
HEADING_LIMIT = 20


def more_heading(state: StateBlock, startLine: int, endLine: int, silent: bool):
    LOGGER.debug("entering heading: %s, %s, %s, %s", state, startLine, endLine, silent)

    pos = state.bMarks[startLine] + state.tShift[startLine]
    maximum = state.eMarks[startLine]

    # if it's indented more than 3 spaces, it should be a code block
    if state.sCount[startLine] - state.blkIndent >= 4:
        return False

    ch: int | None = state.srcCharCode[pos]

    # /* # */
    if ch != 0x23 or pos >= maximum:
        return False

    # count heading level
    level = 1
    pos += 1
    try:
        ch = state.srcCharCode[pos]
    except IndexError:
        ch = None
    # /* # */
    while ch == 0x23 and pos < maximum and level <= HEADING_LIMIT:
        level += 1
        pos += 1
        try:
            ch = state.srcCharCode[pos]
        except IndexError:
            ch = None

    if level > HEADING_LIMIT or (pos < maximum and not isSpace(ch)):
        return False

    if silent:
        return True

    # Let's cut tails like '    ###  ' from the end of string

    maximum = state.skipSpacesBack(maximum, pos)
    tmp = state.skipCharsBack(maximum, 0x23, pos)  # #
    if tmp > pos and isSpace(state.srcCharCode[tmp - 1]):
        maximum = tmp

    state.line = startLine + 1

    token = state.push("heading_open", "h" + str(level), 1)
    token.markup = "".join(islice(repeat("#"), level))
    token.map = [startLine, state.line]

    token = state.push("inline", "", 0)
    token.content = state.src[pos:maximum].strip()
    token.map = [startLine, state.line]
    token.children = []

    token = state.push("heading_close", "h" + str(level), -1)
    token.markup = "".join(islice(repeat("#"), level))

    return True


def replace_heading_rule(md: MarkdownIt):
    md.block.ruler.at("heading", more_heading)
    return md



"""
Utilities for querying various data sources for law texts:
  - legifrance
  - legistix database (standalone)
  - legistix database (auto-api via datasette)
"""

import functools
import logging
from collections.abc import Iterable
from datetime import date, datetime, timedelta, timezone
from typing import Protocol

import aiometer
import httpx
from markdownify import markdownify as md  # type: ignore
from typing_extensions import assert_never

from catleg.config import settings

from catleg.law_text_fr import Article, ArticleType, parse_article_id


def _lf_timestamp_to_datetime(ts):
    return datetime.fromtimestamp(ts / 1000, timezone.utc)


# Legifrance uses 2999-01-01 to mark a non-expired or non-expiring text
END_OF_TIME = _lf_timestamp_to_datetime(32472144000000)


class Backend(Protocol):
    async def article(self, id: str) -> Article | None:
        """
        Retrieve a law article.
        """
        ...

    async def articles(self, ids: Iterable[str]) -> Iterable[Article | None]:
        ...

    async def list_codes(self):
        """
        List available codes.
        """
        ...

    async def code_toc(self, id: str):
        """
        Retrieve the structure of a law code (i.e sections,
        subsections... up to articles).
        Does not include article text.
        """
        ...


class LegifranceBackend(Backend):
    API_BASE_URL = "https://api.piste.gouv.fr/dila/legifrance/lf-engine-app"

    def __init__(self, client_id, client_secret):
        headers = {"Accept": "application/json"}
        self.client = httpx.AsyncClient(
            auth=LegifranceAuth(client_id, client_secret), headers=headers
        )

    async def article(self, id: str) -> Article | None:
        reply = await self.query_article_legi(id)
        article = _article_from_legifrance_reply(reply)
        if article is None:
            logging.warning(f"Could not retrieve article {id} (wrong identifier?)")
        return article

    async def articles(self, ids: Iterable[str]) -> Iterable[Article | None]:
        jobs = [functools.partial(self.query_article_legi, id) for id in ids]
        replies = await aiometer.run_all(jobs, max_at_once=10, max_per_second=15)
        return [_article_from_legifrance_reply(reply) for reply in replies]

    async def list_codes(self):
        return self._list_codes()[0]

    async def _list_codes(self, page_size=20):
        params = {"pageSize": page_size, "pageNumber": 1, "states": ["VIGUEUR"]}
        reply = await self.client.post(f"{self.API_BASE_URL}/list/code", json=params)
        reply.raise_for_status()
        reply_json = reply.json()
        nb_results = reply_json["totalResultNumber"]
        results = reply_json["results"]
        while len(results) < nb_results:
            params["pageNumber"] += 1
            reply = await self.client.post(
                f"{self.API_BASE_URL}/list/code", json=params
            )
            reply.raise_for_status()
            reply_json = reply.json()
            results += reply_json["results"]
        return results, nb_results

    async def code_toc(self, id: str):
        params = {"textId": id, "date": str(date.today())}
        reply = await self.client.post(
            f"{self.API_BASE_URL}/consult/legi/tableMatieres", json=params
        )
        reply.raise_for_status()
        if "sections" not in reply.json():
            raise ValueError(f"Could not retrieve TOC for text {id}")
        return reply.json()

    async def query_article_legi(self, id: str):
        typ, id = parse_article_id(id)
        match typ:
            case ArticleType.LEGIARTI | ArticleType.JORFARTI:
                url = f"{self.API_BASE_URL}/consult/getArticle"
                params = {"id": id}
            case ArticleType.CETATEXT:
                url = f"{self.API_BASE_URL}/consult/juri"
                params = {"textId": id}
            case _:
                raise ValueError("Unknown article type")

        # A POST request to fetch an article?
        # And no way of using a simple query string?
        # Really, Legifrance?
        reply = await self.client.post(url, json=params)
        reply.raise_for_status()
        return reply.json()


def get_backend(spec: str):
    # TODO: multiple backends, fallbacks...
    assert spec == "legifrance"
    client_id, client_secret = _get_legifrance_credentials(raise_if_missing=True)
    return LegifranceBackend(client_id, client_secret)


class LegifranceArticle(Article):
    def __init__(
        self,
        id: str,
        text: str,
        text_html: str,
        nota: str,
        nota_html: str,
        end_date: int | str | None,
        latest_version_id: str,
    ):
        self._id: str = id
        self._text: str = text
        self._text_html: str = text_html
        self._nota: str = nota
        self._nota_html: str = nota_html
        self._end_date: datetime = (
            _lf_timestamp_to_datetime(int(end_date))
            if end_date is not None
            else END_OF_TIME
        )
        self._latest_version_id = latest_version_id

    @property
    def end_date(self) -> datetime:
        return self._end_date

    @property
    def is_open_ended(self) -> bool:
        return self._end_date == END_OF_TIME

    @property
    def id(self) -> str:
        return self._id

    @property
    def text(self) -> str:
        return self._text

    @property
    def text_html(self) -> str:
        return self._text_html

    @property
    def nota(self) -> str:
        return self._nota

    @property
    def nota_html(self) -> str:
        return self._nota_html

    @property
    def latest_version_id(self) -> str:
        return self._latest_version_id

    def text_and_nota(self) -> str:
        if len(self.nota):
            return f"{self._text}\n\nNOTA :\n\n{self._nota}"
        return self._text

    @property
    def type(self) -> ArticleType:
        return parse_article_id(self.id)[0]

    def to_markdown(self) -> str:
        text_md = md(self.text_html, strip=["a"]).strip()
        if len(self.nota_html):
            nota_md = md(self.nota_html, strip=["a"]).strip()
            text_md += f"\n\nNOTA :\n\n{nota_md}"
        return text_md


def _get_legifrance_credentials(*, raise_if_missing=True):
    client_id = settings.get("lf_client_id")
    client_secret = settings.get("lf_client_secret")
    if raise_if_missing:
        if (client_id is None) or (client_secret is None):
            raise ValueError(
                "Please supply Legifrance credentials \
                         (using .catleg_secrets.toml or in the environment)"
            )
    return client_id, client_secret


def _article_from_legifrance_reply(reply) -> Article | None:
    if "article" in reply:
        article = reply["article"]
        if article is None:
            return None
    elif "text" in reply:
        article = reply["text"]
        if article is None:
            return None
    else:
        raise ValueError("Could not parse Legifrance reply")

    article_type, article_id = parse_article_id(article["id"])
    match article_type:
        case ArticleType.CETATEXT:
            latest_version_id = article_id
        case ArticleType.LEGIARTI | ArticleType.JORFARTI:
            article_versions = sorted(
                article["articleVersions"], key=lambda d: int(d["dateDebut"])
            )
            latest_version_id = article_versions[-1]["id"]
        case _:
            assert_never()

    return LegifranceArticle(
        id=article["id"],
        text=article["texte"],
        text_html=article["texteHtml"],
        nota=article["nota"] or "",
        nota_html=article["notaHtml"] or "",
        end_date=article["dateFin"],
        latest_version_id=latest_version_id,
    )


class LegifranceAuth(httpx.Auth):
    """
    Manages authentication for Legifrance API access
    Requires a client id and a client secret, uses those
    to fetch and refresh an access token and passes it along
    in requests.

    Note that the auth flow is synchronous, this is on purpose
    (we do not expect frequent token refreshes).
    See https://github.com/encode/httpx/issues/1176#issuecomment-674381420
    """

    def __init__(self, client_id: str, client_secret: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self.token = None
        self.token_expires_at: datetime | None = None

    def auth_flow(self, request: httpx.Request):
        if self.token is None or self.token_expires_at <= datetime.now():
            logging.info("Requesting auth token")
            data = {
                "grant_type": "client_credentials",
                "scope": "openid",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }
            resp = httpx.post("https://oauth.piste.gouv.fr/api/oauth/token", data=data)
            if not 200 <= resp.status_code < 300:
                yield resp
            resp_json = resp.json()
            self.token = resp_json["access_token"]
            expires_in = int(resp_json["expires_in"])
            self.token_expires_at = datetime.now() + timedelta(seconds=expires_in)
        else:
            logging.info("Using existing auth token")

        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request



import logging
import sys
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import TextIO

from catleg.parse_catala_markdown import parse_catala_file
from catleg.query import get_backend


logger = logging.getLogger(__name__)


async def check_expiry(f: TextIO, *, file_path: Path | None = None):
    # parse articles from file
    articles = parse_catala_file(f, file_path=file_path)

    back = get_backend("legifrance")
    ref_articles = await back.articles([article.id.upper() for article in articles])
    has_expired_articles = False
    now = datetime.now(timezone.utc)

    for article, ref_article in zip(articles, ref_articles):
        if ref_article is None:
            warnings.warn(f"Could not retrieve article '{article.id}'")
            continue

        if article.is_archive:
            logger.info("article '%s' is achived, skipping expiry check", article.id)
            continue

        logger.info("checking article '%s'", article.id)

        if not ref_article.is_open_ended:
            if now > ref_article.end_date:
                print(
                    f"{article.file_path}:{article.start_line} - "
                    f"Article '{article.id}' has expired "
                    f"(on {ref_article.end_date.date()}). "
                    f"It has been replaced by '{ref_article.latest_version_id}'.",
                    file=sys.stderr,
                )
                has_expired_articles = True
            else:
                print(
                    f"{article.file_path}:{article.start_line} - "
                    f"Article '{article.id}' will expire "
                    f"on {ref_article.end_date.date()}. "
                    f"It will be replaced by '{ref_article.latest_version_id}'.",
                    file=sys.stderr,
                )

    return 0 if not has_expired_articles else 1



import re
from dataclasses import dataclass
from datetime import date
from enum import Enum
from pathlib import Path
from typing import Protocol

from more_itertools import intersperse

ArticleType = Enum("ArticleType", ["LEGIARTI", "CETATEXT", "JORFARTI"])


class Article(Protocol):
    @property
    def type(self) -> ArticleType:
        ...

    @property
    def id(self) -> str:
        """Article identifier, for instance LEGIARTI000038814944"""
        ...

    @property
    def text(self) -> str:
        ...


class ExpiryInfo(Protocol):
    @property
    def expiration_date(self) -> date:
        ...


class NewVersionInfo(Protocol):
    @property
    def new_version(self) -> str:
        ...


class FileLineInfo(Protocol):
    @property
    def file_path(self) -> Path | None:
        ...

    @property
    def start_line(self) -> int:
        ...


@dataclass(frozen=True)
class CatalaFileArticle:
    type: ArticleType
    id: str
    text: str
    file_path: Path | None
    start_line: int
    is_archive: bool


def parse_article_id(article_id: str) -> tuple[ArticleType, str]:
    """
    Parse a string that looks like 'article-l822-2-legiarti000038814944'
    and return (ArticleType.LEGIARTI, 'legiarti000038814944')
    """
    article_id = article_id.split("-")[-1]
    typ = ArticleType[article_id[:8].upper()]
    return typ, article_id


# Some people, when confronted with a problem, think
# "I know, I'll use regular expressions."
# Now they have two problems. (Jamie Zawinsky)
TYPES_OR_REGEX = "".join(intersperse("|", [typ.name for typ in ArticleType]))
ARTICLE_ID_REGEX = re.compile(f"\\b(({TYPES_OR_REGEX})[0-9]{{12}})\\b")


def find_id_in_string(text: str) -> tuple[ArticleType, str] | None:
    """
    Finds an article ID in a string, returns a tuple (ArticleType, ID)
    if found or None if the string does not contain an article ID.
    """
    match = ARTICLE_ID_REGEX.search(text)
    if match:
        return parse_article_id(match.group(0))
    return None
